/* formant.c -- copyright 1994 Stefan Bilbao */
/* ported to Pd by MSP 2003.  Also added decimation argument */

static char formant_version[] = "formant version 0.2";

#include <math.h>
#include "m_pd.h"

#define MINPOINTS 128
#define MAXPOINTS 16384
#define WMAX 40
#define VECTOR_SIZE 64
#define SPEC_SIZE 128
#define ONE_TWELFTH 0.083333
#define PI 3.14159           
#define SPEC_SIZE_MINUS_ONE 127
#define RAW_FORM_NUM 4      /* number of raw formants detected */
#define FORM_NUM 3      /* number of continuously tracked ones */

static int printnext;
#define S_OFF 0
#define S_ON 1
#define S_NEXTON 2


typedef struct peak
{
    float p_freq;
    float p_bw;
    float p_db;
    int p_status;
} t_peak;

typedef struct ctlformant
{
    float *inbuf_start;
    float *inbuf_end;
    float *inbuf_current;
    double *currentOutBuf;
    double *prevOutBuf;
    double *outbufA;
    double *outbufB;
    double *finalend;
    double *finalcurrent;
    double *sinboink;
    double *cosboink;
    t_peak *formantout;
    float rawformant[RAW_FORM_NUM];
    void *clk;
    int uzi;        /* flag to put results out every period */
    float maxfreq;  /* max. formant frequency I care about */
    float samplerate;
} t_ctlformant;

typedef struct formant
{
    t_object x_obj;
    t_outlet *x_out[5]; 
    t_ctlformant *x_ctl;
    t_ctlformant x_cspace;
    int x_npts;
    int x_wnum;
    int x_numinterp;
    int x_index;
    int x_decimation;
    int x_decphase;
    float debug;
} t_formant;

static void funcopy(double *f, double *g, int n)
{
    double *new, *old;
    new = g;
    old = f;
    while (n--) *new++ = *old++;
}

static void funzero(double *f, int n)
{
    double *dorky;
    
    dorky = f;
    while (n--) *dorky++ = 0.;
}       

static void formant_clear(t_formant *x)
{
    int i;
    for (i = 0; i < FORM_NUM; i++) x->x_ctl->formantout[i].p_status = 0;
}

static void formant_maxfreq(t_formant *x, t_floatarg f)
{
    x->x_ctl->maxfreq = f;
}

static void formant_bang(t_formant *x)
{
    int i, n = FORM_NUM;
    t_peak *foo = x->x_ctl->formantout;
    t_atom form[RAW_FORM_NUM > 3 ? RAW_FORM_NUM : 3];
    t_outlet **op;

    for (i = 0; i < RAW_FORM_NUM; i++)
    {
        form[i].a_type = A_FLOAT;
        /*
        form[i].a_w.w_long = (x->x_ctl->rawformant[i] > 0 ?
            69. + 17.31234 * log(x->x_ctl->rawformant[i]*(1./440.)) : 0); */
        form[i].a_w.w_float = x->x_ctl->rawformant[i];
    }
    outlet_list(x->x_out[4], (void *)0, RAW_FORM_NUM, form);
    op  = &x->x_out[1];
    form[0].a_type = A_FLOAT;
    form[1].a_type = form[2].a_type = A_FLOAT;

    for (i = 0; i < n; i++, foo++)
    {
        float freq = foo->p_freq;
        float amp = foo->p_db;
        int valid = (foo->p_status == S_ON);
        form[0].a_w.w_float = valid;
        form[1].a_w.w_float = freq;
        form[2].a_w.w_float = amp;
        outlet_list(*op++, (void *)0, 3, form);
    }
    outlet_bang(x->x_out[0]);
}

static void formant_print(t_formant *x)
{
    post("formant: max freq %ld", (long)(x->x_ctl->maxfreq));
    printnext = 1;
}

static int snoodle;

#define MINPEAK 0

static void LPC(t_formant *x)
{
    t_ctlformant *c = x->x_ctl;
    double *a1, *a2, *p1, *p2, *p3, *u1, *u2, *u3, *u4, *zero,
        *temp, *doinker, *z, *zend, *re, *s;
    double AUTO[WMAX+1], BIG_E, SUM, K, ALPHA[40][40], size, gain,
        reform[SPEC_SIZE], imform[SPEC_SIZE], power[SPEC_SIZE], slope,
        lastslope, concav, lastconcav, dBpow, dBpow1,
        dBpow2, para_k, para_h, para_cf;    
    
    t_peak DERIV_FIT[RAW_FORM_NUM];
    int i, i1, W, zcnum = 0, t, t2, j, m, index1 = 0; 
    int n, w; 

    n = x->x_npts;
    w = x->x_wnum;
    /*post("%f \n", *c->inbuf_start);*/
    funzero(&AUTO[0], w+1);
    a1 = &AUTO[0];
    a2 = &AUTO[w];
    while(a1 <= a2)   /* calculate the autocorrelations   */
    {
        float *fp1, *fp2, *fp3;
        double foo;
        double gomper;
        snoodle = 1;
        fp1 = c->inbuf_start+w-(a2-a1);
        fp2 = c->inbuf_end;
        fp3 = c->inbuf_start;
#if 0
        dotproduct(fp1, fp3, &foo, fp2-fp1);
        *a1++ = foo;
#elif 0
        gomper = 0;
        while(fp1 < fp2)
        {
            gomper += ((double)(*fp1)*(double)(*fp3));
            /*post("%f %f %f \n", *p1, *p3, *a1);*/
        
            fp1++;
            fp3++;
        }
        *a1++ = gomper;
#else
    	for (i = 0, gomper = 0; i < n - w - 1; i++)
            gomper += ((double)(*fp1++)*(double)(*fp3++));
        *a1++ = gomper;
#endif
    }
    /* if (printnext)
    {
        int xx;
        post("order %d", w);
        for (xx = 0; xx < w; xx += 4)
        {
            post("%f\t%f\t%f\t%f", 
            (float)AUTO[xx],(float)AUTO[xx+1],
            (float)AUTO[xx+2],(float)AUTO[xx+3]);
        }
        post("%f", (float)AUTO[w]);
    } */
    BIG_E = AUTO[0];
    /*Durbin's recursive algorithm*/ 
    for(i=1;i<=w;i++)   
    {
        i1 = i-1;
        SUM = 0.;
        u1 = &ALPHA[i1][1];
        u2 = &ALPHA[i1][i];
        a1 = &AUTO[i1];
        while(u1 < u2)
        {
            SUM = SUM + ((*u1)*(*a1));
            u1++;
            a1--;
        }
        K = (AUTO[i]-SUM)/BIG_E;
        ALPHA[i][i] = K;
        u1 = &ALPHA[i][1];
        u2 = &ALPHA[i1][1];
        u3 = &ALPHA[i1][i1];
        u4 = &ALPHA[i][i];
        while(u1 < u4)
        {
            *u1 = *u2 - (K*(*u3));
            u1++;
            u2++;
            u3--;
        }
        BIG_E = (1-(K*K))*BIG_E;
        /* if (printnext)
        {
            post("SUM %f, BIG_E %f, K %f", (float)SUM,
                (float)BIG_E, (float)K);
            if (i == 1) post("- %f", ALPHA[1][1]);
            if (i == 2) post("- %f %f", ALPHA[2][1], ALPHA[2][2]);
            if (i == 3) post("- %f %f %f", ALPHA[3][1], ALPHA[3][2],
                ALPHA[3][3]);
            if (i == 4) post("- %f %f %f %f", ALPHA[4][1], ALPHA[4][2],
                ALPHA[4][3], ALPHA[4][4]);
        } */
    }
    gain = AUTO[0];
    for(i=1;i<=w;i++)
    {
        gain = gain + ALPHA[w][i]*AUTO[i];
    }
    /*z = c->inbuf_start;
    zend = c->inbuf_end;
    zcnum = 0;
    while(z < zend)
    {
        if(((*z)*(*(z-1)))<=0)
        {
            zcnum = zcnum + 1;
        }
        z++;
    }*/
    /************begin formant fitting from lpc data****************/
    
    /*********draw lpc spectrum, log the bins*******************/
    /* if (printnext)
    {
        int xx;
        for (xx = 0; xx < w; xx += 4)
        {
            post("%f\t%f\t%f\t%f", 
            (float)ALPHA[w][xx+1],(float)ALPHA[w][xx+2],
            (float)ALPHA[w][xx+2],(float)ALPHA[w][xx+3]);
        }
    } */
    
    re = c->cosboink;
    s = c->sinboink;
    for(t=0;t<SPEC_SIZE;t++)
    {
        reform[t] = 1.;
        imform[t] = 0.;
        doinker = &ALPHA[w][1];
        j = w;
        while(j--)
        {
            reform[t] = reform[t] - (*doinker)*(*re++); 
            imform[t] = imform[t] + (*doinker++)*(*s++);
        }
        power[t] = (1./((pow(reform[t],2)+pow(imform[t],2))));
    }
    /*
    if (printnext)
    {
        int xx;
        for (xx = 0; xx < 128; xx += 4)
        {
            post("%f %f %f %f",
                (float)(power[xx]), 
                (float)(power[xx+1]), 
                (float)(power[xx+2]), 
                (float)(power[xx+3]));
        }
    }
    */
    /*************find peaks by using derivative tests***********/
    
    lastslope = power[2]-power[1];
    lastconcav = power[2]-2*power[1]+power[0];
    m = 0;
    for(t=3;t<SPEC_SIZE;t++)
    {
        if (m < RAW_FORM_NUM)
        {
            float newfreq;
            slope = power[t] - power[t-1];
            concav = slope-lastslope;
            if((slope*lastslope)<0)
            {   
                if(concav<0)
                {
                    t2 = t-2;
                    dBpow = 10*log(power[t]);
                    dBpow1 = 10*log(power[t-1]);
                    dBpow2 = 10*log(power[t2]);      
                    para_cf=t2+((-3*dBpow2+4*dBpow1-dBpow)
                    /(-2*dBpow2+4*dBpow1-2*dBpow));
                    para_k = (dBpow1-dBpow2)/(2*t-3-2*para_cf);
                    /* para_h = dBpow2-para_k*(t2-para_cf)*(t2-para_cf); */
                    para_h = dBpow1;
                    DERIV_FIT[m].p_freq = newfreq = pow(2, 
                    (para_cf-SPEC_SIZE+1)/12)*(c->samplerate/2.);
                    DERIV_FIT[m].p_db= para_h;
                    DERIV_FIT[m].p_bw = -2/concav;
                    if (para_h > MINPEAK && newfreq < c->maxfreq)
                    {
                        if (printnext) post("%10.6d\t%d",
                            (int)(DERIV_FIT[m].p_freq),
                            (int)(DERIV_FIT[m].p_db));
                        m++;
                    }
                    index1 = 1;
                }
            }
            if((concav*lastconcav)<0)
            {
                t2 = t-2;
                dBpow = 10*log(power[t]);
                dBpow1 = 10*log(power[t-1]);
                dBpow2 = 10*log(power[t2]);
                if((((slope>0) && 
                ((concav-lastconcav)>0)) || ((slope<0) && 
                ((concav-lastconcav)<0))) && (index1 == 0))
                {
                    DERIV_FIT[m].p_freq = newfreq = pow(2.,
                        (t-SPEC_SIZE_MINUS_ONE)*ONE_TWELFTH)*
			    (c->samplerate/2.);
                    DERIV_FIT[m].p_db = dBpow;
                    DERIV_FIT[m].p_bw = -0.05;
                    if (dBpow > MINPEAK && newfreq < c->maxfreq)
                    {
                        if (printnext) post("%10.6d\t%d",
                            (int)(DERIV_FIT[m].p_freq),
                            (int)(DERIV_FIT[m].p_db));
                        m++;
                    }
                }
            }
            lastslope = slope;
            lastconcav = concav; 
            index1 = 0.;
        }
    }

/****************end formant fitter*******************************/ 
    
        /* save raw LPC data -- not used in current version */
    *(c->currentOutBuf + 60) = (double) zcnum;
    *c->currentOutBuf = sqrt(gain);
    funcopy(&ALPHA[w][0]+1, c->currentOutBuf+1, w);

        /* try to line up the peaks with the previous ones. */

    for (i = 0; i < m; i++) c->rawformant[i] = DERIV_FIT[i].p_freq;
    for (i = m; i < RAW_FORM_NUM; i++) c->rawformant[i] = 0;
    if (printnext)
    {
        for (i = 0; i < 4; i++)
            post("formant %d ...", i); postfloat(c->rawformant[i]);
    }

    if (m > FORM_NUM) m = FORM_NUM; /* drop extra formants */
    for (i = 0; i < m; i++)
    {
        t_peak *peak1 = &DERIV_FIT[i];
        t_peak *peak2 = c->formantout;
        peak1->p_status = -1;
        for (j = 0; j < FORM_NUM; j++, peak2++)
        {
            if (peak2->p_status && peak2->p_freq < peak1->p_freq * 1.2 &&
                peak2->p_freq > peak1->p_freq * .8)
            {
                peak1->p_status = j;
                break;
            }
        }
    }

        /* initialize current states to zero */
    for (i = 0; i < FORM_NUM; i++) c->formantout[i].p_status = 0;

        /* mark states of current formants that match a new one as ON */
    for (i = 0; i < m; i++) if (DERIV_FIT[i].p_status >= 0)
        c->formantout[DERIV_FIT[i].p_status].p_status = S_ON;

            /* for new formants that didn't get a match, find a free
                one and mark it "NEXTON". */
    for (i = 0, j = 0; i < m; i++)
    {
        if (DERIV_FIT[i].p_status < 0)
        {
            while (c->formantout[j].p_status)
            {
                if (j >= FORM_NUM)
                {
                    post("bug: lpc 1");
                    j = 0;
                    break;
                }
                j++;
            }
            DERIV_FIT[i].p_status = j;
            c->formantout[j].p_status = S_NEXTON;
        }
    }
        /* copy new ones in.  Old ones that aren't copied onto are already
            marked "off".  */
    for (i = 0; i < m; i++) 
    {
        t_peak *peak1 = &DERIV_FIT[i];
        t_peak *peak2 = c->formantout + peak1->p_status;
        if (peak1->p_status >= 0)
        {
            peak2->p_freq = peak1->p_freq;
            peak2->p_db = peak1->p_db;
            peak2->p_bw = peak1->p_bw;
        }
        else post("LPC bug 2");
    }

    /* if (printnext)
    {
        for (i = 0; i < FORM_NUM; i++)
            if (c->formantout[i].p_status)
                post("%d ---> %8.6d %d", i, (int)c->formantout[i].p_freq, 
                    c->formantout[i].p_status);
            else post("%d off", i);
    } */
    printnext = 0;
    if (c->uzi) clock_delay(c->clk, 0);
}

static t_int *formant_perform(t_int *w)
{
    t_float *in1 = (t_float *)(w[1]);
    t_formant *x = (t_formant *)(w[2]);
    int n = (int)(w[3]);
    int decphase = x->x_decphase, decimation = x->x_decimation;

    int flag, here, vl;
    t_ctlformant *c = x->x_ctl;
    while (n--) 
    {
        
        if (c->inbuf_current == c->inbuf_end)   
        {
            if(c->currentOutBuf == c->outbufB)
            {
                c->currentOutBuf = c->outbufA;
                c->prevOutBuf = c->outbufB;
            }
            else
            {
                c->currentOutBuf = c->outbufB;
                c->prevOutBuf = c->outbufA;
            }
            LPC(x); 
            c->inbuf_current = c->inbuf_start;
            x->x_index = 0;
        
        }
	if (!decphase)
	{
            *c->inbuf_current++ = *in1;
	    decphase = decimation;
	}
	in1++;
	decphase--;
    }
    x->x_index++;
    x->x_decphase = decphase;
    return (w+4);
}

static void formant_gentab(t_formant *x)
{
    /*********generate sines and cosines for spectrum drawing**************/
    double samplerate = x->x_ctl->samplerate;
    double shit = samplerate/(2.*PI), boink, jboink;
    double *SIN = x->x_ctl->sinboink;
    double *COS = x->x_ctl->cosboink;
    int j, t;
    int w = x->x_wnum;
    for(t=0;t<SPEC_SIZE;t++)
    {
        boink = PI*pow(2., (t-SPEC_SIZE_MINUS_ONE)*ONE_TWELFTH);
        for(j=1;j<=w;j++)
        {
            jboink = j*boink;
            *SIN++ = sin(jboink);
            *COS++ = cos(jboink);
        }
    }
}

static void formant_dsp(t_formant *x, t_signal **sp)
{
    if (x->x_ctl->samplerate != sp[0]->s_sr / x->x_decimation)
    {
    	x->x_ctl->samplerate = sp[0]->s_sr / x->x_decimation;
	formant_gentab(x);
    }
    dsp_add(formant_perform, 3, sp[0]->s_vec, x, sp[0]->s_n);
}

static void formant_uzi(t_formant *x, t_floatarg f)
{
    x->x_ctl->uzi = (f != 0);
}

static t_class *formant_class;

static int funlog2(int n)
{
    int ret = -1;
    while (n)
    {
        n >>= 1;
        ret++;
    }
    return (ret);
}

void *formant_new(t_floatarg fnpoints, t_floatarg fw, t_floatarg fdec)
{
    register t_formant *x;
    double *fp, jboink;
    int logn, i, t, j;
    int npoints = fnpoints;
    int w = fw;
    int decimation = fdec;
    if (decimation <= 0)
    	decimation = 1;
    if (!npoints) npoints = 512;
    if (!w) w = 20;
    logn = funlog2(npoints);
    if (npoints != (1 << logn))
    {
        post("formant: power of two required as argument");
        return (0);
    }
    if (npoints > MAXPOINTS || npoints < MINPOINTS)
    {
        post("formant: argument out of range");
        return (0);
    }
    if (w > WMAX || w < 1)
    {
        error("w: out of range");
        return (0);
    }
    x = (t_formant *)pd_new(formant_class);

    x->x_out[0] = outlet_new(&x->x_obj, gensym("bang"));
    x->x_out[1] = outlet_new(&x->x_obj, gensym("list"));
    x->x_out[2] = outlet_new(&x->x_obj, gensym("list"));
    x->x_out[3] = outlet_new(&x->x_obj, gensym("list"));
    x->x_out[4] = outlet_new(&x->x_obj, gensym("list"));
    
    x->x_ctl = &x->x_cspace;
    x->x_ctl->inbuf_start = (float *)getbytes((npoints)*sizeof(float));
    x->x_ctl->inbuf_current = x->x_ctl->inbuf_start;
    x->x_ctl->inbuf_end = x->x_ctl->inbuf_start + npoints;
    x->x_wnum = w;
    x->x_npts = npoints;
    x->x_numinterp = npoints/VECTOR_SIZE; 
    x->x_index = 0;
    x->x_ctl->outbufA = (double *)getbytes((VECTOR_SIZE)*sizeof(double)); 
    x->x_ctl->outbufB = (double *)getbytes((VECTOR_SIZE)*sizeof(double)); 
    x->x_ctl->formantout = (t_peak *)getbytes((FORM_NUM)*sizeof(t_peak));
    x->x_ctl->sinboink = (double *)getbytes((w*SPEC_SIZE)*sizeof(double)); 
    x->x_ctl->cosboink = (double *)getbytes((w*SPEC_SIZE)*sizeof(double)); 
    funzero(x->x_ctl->outbufA, VECTOR_SIZE);
    x->x_ctl->currentOutBuf = x->x_ctl->outbufA;
    x->x_ctl->prevOutBuf = x->x_ctl->outbufB;
    x->x_ctl->uzi = 0;
    x->x_ctl->clk = clock_new(x, (t_method)formant_bang);
    x->x_ctl->maxfreq = 1000000.;
    x->debug = 0.;
    x->x_ctl->samplerate = 1;
    x->x_decimation = decimation;
    x->x_decphase = 0;
    formant_gentab(x);
    formant_clear(x);
    return (x);
}

void formant_ff(t_formant *x)		/* cleanup on free */
{
    freebytes(x->x_ctl->inbuf_start, x->x_npts * sizeof (float));
    freebytes(x->x_ctl->outbufA , VECTOR_SIZE * sizeof (double));
    freebytes(x->x_ctl->outbufB , VECTOR_SIZE * sizeof (double));
    freebytes(x->x_ctl->formantout , FORM_NUM * sizeof (double));
    freebytes(x->x_ctl->sinboink , (x->x_wnum*SPEC_SIZE) * sizeof (double));
    freebytes(x->x_ctl->cosboink , (x->x_wnum*SPEC_SIZE) * sizeof (double));
    clock_free(x->x_ctl->clk);
}


void formant_setup(void)
{
    formant_class = class_new(gensym("formant"), (t_newmethod)formant_new,
    	(t_method)formant_ff, sizeof(t_formant), 0,
    	    A_DEFFLOAT, A_DEFFLOAT, A_DEFFLOAT, 0);
    class_addmethod(formant_class, (t_method)formant_dsp,
    	gensym("dsp"), 0);
    class_addmethod(formant_class, nullfn, gensym("signal"), 0);
    class_addbang(formant_class, formant_bang);
    class_addmethod(formant_class, (t_method)formant_print,
    	gensym("print"), 0);
    class_addmethod(formant_class, (t_method)formant_clear,
    	gensym("clear"), 0);
    class_addmethod(formant_class, (t_method)formant_uzi,
    	gensym("uzi"), A_FLOAT, 0);
    class_addmethod(formant_class, (t_method)formant_maxfreq,
    	gensym("maxfreq"), A_FLOAT, 0);

    post(formant_version);
}


