/* paf4x -- version of the paf generator for en echo, which reproduces a
bug in teh old 4X version -- search for "4XBUG" */

/*  Copyright 1997-1999 Miller Puckette.
Permission is granted to use this software for any purpose provided you
keep this copyright notice intact.

THE AUTHOR AND HIS EMPLOYERS MAKE NO WARRANTY, EXPRESS OR IMPLIED,
IN CONNECTION WITH THIS SOFTWARE.

This file is downloadable from http://www.crca.ucsd.edu/~msp .

Warning: the algorithms implemented here are covered by patents held by
IRCAM.  While this probably does not restrict anyone from distributing
software implementing the paf4x, any hardware implementor should obtain a
license from IRCAM.
*/

static char *paf4x_version = "paf4x version 0.06";

#ifdef NT
#pragma warning (disable: 4305 4244)
#endif

#ifdef TESTME
#include <stdio.h>
#endif
#ifdef FTS15
#include "mess.h"
#include "dsp.h"
#endif
#ifdef V26SGI
#include "m_extern.h"
#include "d_graph.h"
#include "d_ugen.h"
#endif
#ifdef PD
#include "m_pd.h"
#endif
#include <math.h>

#define LOGTABSIZE 9
#define TABSIZE (1 << LOGTABSIZE)
#define TABRANGE 3

typedef struct _tabpoint
{
    float p_y;
    float p_diff;
} t_tabpoint;

static t_tabpoint paf4x_gauss[TABSIZE];
static t_tabpoint paf4x_cauchy[TABSIZE];

typedef struct _linenv
{
    double l_current;
    double l_biginc;
    float l_1overn;
    float l_target;
    float l_msectodsptick;
    int l_ticks;
} t_linenv;

typedef struct _paf4xctl
{
    t_linenv x_freqenv;
    t_linenv x_cfenv;
    t_linenv x_bwenv;
    t_linenv x_ampenv;
    t_linenv x_vibenv;
    t_linenv x_vfrenv;
    t_linenv x_shiftenv;
    float x_isr;
    float x_held_freq;
    float x_held_intcar;
    float x_held_fraccar;
    float x_held_bwquotient;
    double x_phase;
    double x_shiftphase;
    double x_vibphase;
    int x_triggerme;
    int x_cauchy;
    int x_4xcompat;
} t_paf4xctl;

static void linenv_debug(t_linenv *l, char *s)
{
#ifdef DEBUG
    post("%s: current %f, target %f", s, l->l_current, l->l_target);
#endif
}

static void linenv_init(t_linenv *l)
{
    l->l_current = l->l_biginc = 0;
    l->l_1overn = l->l_target = l->l_msectodsptick = 0;
    l->l_ticks = 0;
}

static void linenv_setsr(t_linenv *l, float sr, int vecsize)
{
    l->l_msectodsptick = sr / (1000.f * ((float)vecsize));
    l->l_1overn = 1.f/(float)vecsize;
}

static void linenv_set(t_linenv *l, float target, long timdel)
{
    if (timdel > 0)
    {
	l->l_ticks = ((float)timdel) * l->l_msectodsptick;
	if (!l->l_ticks) l->l_ticks = 1;
	l->l_target = target;
	l->l_biginc = (l->l_target - l->l_current)/l->l_ticks;
    }
    else
    {
	l->l_ticks = 0;
	l->l_current = l->l_target = target;
	l->l_biginc = 0;
    }
}

#define LINENV_RUN(linenv, current, incr) 		\
    if (linenv.l_ticks > 0)				\
    {							\
    	current = linenv.l_current;			\
    	incr = linenv.l_biginc * linenv.l_1overn;	\
	linenv.l_ticks--;				\
	linenv.l_current += linenv.l_biginc;		\
    }							\
    else						\
    {							\
    	linenv.l_current = current = linenv.l_target;	\
	incr = 0;					\
    }

#define UNITBIT32 1572864.		/* 3*2^19 -- bit 32 has value 1 */
#define TABFRACSHIFT (UNITBIT32/TABSIZE) 

#if defined(__FreeBSD__) || defined(__APPLE__) || defined(__FreeBSD_kernel__)
#include <machine/endian.h>
#endif

#if defined(__linux__) || defined(__CYGWIN__) || defined(__GNU__) || defined(ANDROID)
#include <endian.h>
#endif

#ifdef __MINGW32__
#include <sys/param.h>
#endif

#ifdef _MSC_VER
/* _MSVC lacks BYTE_ORDER and LITTLE_ENDIAN */
#define LITTLE_ENDIAN 0x0001
#define BYTE_ORDER LITTLE_ENDIAN
#endif

#if !defined(BYTE_ORDER) || !defined(LITTLE_ENDIAN)                         
#error No byte order defined                                                    
#endif

#if BYTE_ORDER == LITTLE_ENDIAN
# define HIOFFSET 1                                                              
# define LOWOFFSET 0                                                             
#else                                                                           
# define HIOFFSET 0    /* word offset to find MSB */                             
# define LOWOFFSET 1    /* word offset to find LSB */                            
#endif

#ifdef _MSC_VER
 typedef __int32 int32_t; /* use MSVC's internal type */
#elif defined(IRIX)
 typedef long int32_t;  /* a data type that has 32 bits */
#else
# include <stdint.h>  /* this is where int32_t is defined in C99 */
#endif

#define A1 ((float)(4 * (3.14159265/2)))
#define A3 ((float)(64 * (2.5 - 3.14159265)))
#define A5 ((float)(1024 * ((3.14159265/2) - 1.5)))

static void paf4xctl_run(t_paf4xctl *x, float *out1, int n)
{
    float freqval, freqinc;
    float cfval, cfinc;
    float bwval, bwinc;
    float ampval, ampinc;
    float vibval, vibinc;
    float vfrval, vfrinc;
    float shiftval, shiftinc;
    float bwquotient, bwqincr;
    double  ub32 = UNITBIT32;
    double phase = x->x_phase + ub32;
    double phasehack;
    volatile double *phasehackp = &phasehack;
    double shiftphase = x->x_shiftphase + ub32;
    volatile int32_t *hackptr = ((int32_t *)(phasehackp)) + HIOFFSET, hackval;
    volatile int32_t *lowptr = ((int32_t *)(phasehackp)) + LOWOFFSET, lowbits;
    float held_freq = x->x_held_freq;
    float held_intcar = x->x_held_intcar;
    float held_fraccar = x->x_held_fraccar;
    float held_bwquotient = x->x_held_bwquotient;
    float sinvib, vibphase;
    t_tabpoint *paf4x_table = (x->x_cauchy ? paf4x_cauchy : paf4x_gauss);
    *phasehackp = ub32;
    hackval = *hackptr;

    	/* fractional part of shift phase */
    *phasehackp = shiftphase;
    *hackptr = hackval;
    shiftphase = *phasehackp;

	/* propagate line envelopes */
    LINENV_RUN(x->x_freqenv, freqval, freqinc);
    LINENV_RUN(x->x_cfenv, cfval, cfinc);
    LINENV_RUN(x->x_bwenv, bwval, bwinc);
    LINENV_RUN(x->x_ampenv, ampval, ampinc);
    LINENV_RUN(x->x_vibenv, vibval, vibinc);
    LINENV_RUN(x->x_vfrenv, vfrval, vfrinc);
    LINENV_RUN(x->x_shiftenv, shiftval, shiftinc);

	/* fake line envelope for quotient of bw and frequency */
    bwquotient = bwval/freqval;
    bwqincr = (((float)(x->x_bwenv.l_current))/
    	((float)(x->x_freqenv.l_current)) - bwquotient) *
		x->x_freqenv.l_1overn;

	/* run the vibrato oscillator */
    
    *phasehackp = ub32 + (x->x_vibphase + n * x->x_isr * vfrval);
    *hackptr = hackval;
    vibphase = (x->x_vibphase = *phasehackp - ub32);
    if (vibphase > 0.5)
    	sinvib = 1.0f - 16.0f * (0.75f-vibphase) * (0.75f - vibphase);
    else sinvib = -1.0f + 16.0f * (0.25f-vibphase) * (0.25f - vibphase);
    freqval = freqval * (1.0f + vibval * sinvib);

    shiftval *= x->x_isr;
    shiftinc *= x->x_isr;
    
	/* if phase or amplitude is zero, load in new params */
    if (ampval == 0 || phase == ub32 || x->x_triggerme)
    {
    	    /* 4XBUG -- this should be just cfval/freqval: */
    	float cf_over_freq =
	    (x->x_4xcompat ? 2 * cfval/freqval - 1 : cfval/freqval);
    	held_freq = freqval * x->x_isr;
	held_intcar = (float)((int)cf_over_freq);
	held_fraccar = cf_over_freq - held_intcar;
	held_bwquotient = bwquotient;
	x->x_triggerme = 0;
    }
    while (n--)
    {
    	double newphase = phase + held_freq;
	double carphase1, carphase2, fracnewphase;
	float fphase, fcarphase1, fcarphase2, carrier;
	float g, g2, g3, cosine1, cosine2, halfsine, mod, tabfrac;
	t_tabpoint *p;
   	    /* put new phase into 64-bit memory location.  Bash upper
	    32 bits to get fractional part (plus "ub32").  */

	*phasehackp = newphase;
	*hackptr = hackval;
	newphase = *phasehackp;
	fracnewphase = newphase-ub32;
	fphase = 2.0f * ((float)(fracnewphase)) - 1.0f;
	if (newphase < phase)
	{
	    float cf_over_freq =    	/* 4XBUG */
	    	(x->x_4xcompat ? 2 * cfval/freqval - 1 : cfval/freqval);
	    held_freq = freqval * x->x_isr;
	    held_intcar = (float)((int)cf_over_freq);
	    held_fraccar = cf_over_freq - held_intcar;
	    held_bwquotient = bwquotient;
	}
	phase = newphase;
	*phasehackp = fracnewphase * held_intcar + shiftphase;
	*hackptr = hackval;
	carphase1 = *phasehackp;
	fcarphase1 = carphase1 - ub32;
	*phasehackp = carphase1 + fracnewphase;
	*hackptr = hackval;
	carphase2 = *phasehackp;
	fcarphase2 = carphase2 - ub32;
    	
	shiftphase += shiftval;
	
    	if (fcarphase1 > 0.5f)  g = fcarphase1 - 0.75f;
    	else g = 0.25f - fcarphase1;
    	g2 = g * g;
    	g3 = g * g2;
    	cosine1 = g * A1 + g3 * A3 + g2 * g3 * A5;

    	if (fcarphase2 > 0.5f)  g = fcarphase2 - 0.75f;
    	else g = 0.25f - fcarphase2;
    	g2 = g * g;
    	g3 = g * g2;
    	cosine2 = g * A1 + g3 * A3 + g2 * g3 * A5;
	
	carrier = cosine1 + held_fraccar * (cosine2-cosine1);

    	ampval += ampinc;
	bwquotient += bwqincr;

	/* printf("bwquotient %f\n", bwquotient); */

	halfsine = held_bwquotient * (1.0f - fphase * fphase);
	if (halfsine >= (float)(0.997 * TABRANGE))
	    halfsine = (float)(0.997 * TABRANGE);

#if 0
    	shape = halfsine * halfsine;
	mod = ampval * carrier *
	    (1 - bluntval * shape) / (1 + (1 - bluntval) * shape);
#endif
#if 0
    	shape = halfsine * halfsine;
	mod = ampval * carrier *
	    exp(-shape);
#endif
	halfsine *= (float)(1./TABRANGE);

   	    /* Get table index for "halfsine".  Bash upper
	    32 bits to get fractional part (plus "ub32").  Also grab
	    fractional part as a fixed-point number to use as table
	    address later. */

	*phasehackp = halfsine + ub32;
	lowbits = *lowptr;

    	    /* now shift again so that the fractional table address
	    appears in the low 32 bits, bash again, and extract this as
	    a floating point number from 0 to 1. */
	*phasehackp = halfsine + TABFRACSHIFT;
	*hackptr = hackval;
    	tabfrac = *phasehackp - ub32;

    	p = paf4x_table + (lowbits >> (32 - LOGTABSIZE));
	mod = ampval * carrier * (p->p_y + tabfrac * p->p_diff);
	
	*out1++ = mod;
    }
    x->x_phase = phase - ub32;
    x->x_shiftphase = shiftphase - ub32;
    x->x_held_freq = held_freq;
    x->x_held_intcar = held_intcar;
    x->x_held_fraccar = held_fraccar;
    x->x_held_bwquotient = held_bwquotient;
}

static void paf4xctl_init(t_paf4xctl *x)
{
    linenv_init(&x->x_freqenv);
    linenv_init(&x->x_cfenv);
    linenv_init(&x->x_bwenv);
    linenv_init(&x->x_ampenv);
    linenv_init(&x->x_vibenv);
    linenv_init(&x->x_vfrenv);
    linenv_init(&x->x_shiftenv);
    x->x_freqenv.l_target = x->x_freqenv.l_current = 1.0;
    x->x_isr = (float)(1./44100.);
    x->x_held_freq = 1.f;
    x->x_held_intcar = 0.f;
    x->x_held_fraccar = 0.f;
    x->x_held_bwquotient = 0.f;
    x->x_phase = 0.;
    x->x_shiftphase = 0.;
    x->x_vibphase = 0.;
    x->x_triggerme = 0;
    x->x_cauchy = 1;     /* 4XBUG -- cauchy is 1 by default. */
}

static void paf4xctl_setsr(t_paf4xctl *x, float sr, int vecsize)
{
    x->x_isr = 1.f/sr;
    linenv_setsr(&x->x_freqenv, sr, vecsize);
    linenv_setsr(&x->x_cfenv, sr, vecsize);
    linenv_setsr(&x->x_bwenv, sr, vecsize);
    linenv_setsr(&x->x_ampenv, sr, vecsize);
    linenv_setsr(&x->x_vibenv, sr, vecsize);
    linenv_setsr(&x->x_vfrenv, sr, vecsize);
    linenv_setsr(&x->x_shiftenv, sr, vecsize);
}

static void paf4xctl_freq(t_paf4xctl *x, float val, int time)
{
    if (val < 1.f) val = 1.f;
    if (val > 10000000.f) val = 1000000.f;
    linenv_set(&x->x_freqenv, val, time);
}

static void paf4xctl_cf(t_paf4xctl *x, float val, int time)
{
    linenv_set(&x->x_cfenv, val, time);
}

static void paf4xctl_bw(t_paf4xctl *x, float val, int time)
{
    linenv_set(&x->x_bwenv, val, time);
}

static void paf4xctl_amp(t_paf4xctl *x, float val, int time)
{
    linenv_set(&x->x_ampenv, val, time);
}

static void paf4xctl_vib(t_paf4xctl *x, float val, int time)
{
    linenv_set(&x->x_vibenv, val, time);
}

static void paf4xctl_vfr(t_paf4xctl *x, float val, int time)
{
    linenv_set(&x->x_vfrenv, val, time);
}

static void paf4xctl_shift(t_paf4xctl *x, float val, int time)
{
    linenv_set(&x->x_shiftenv, val, time);
}

static void paf4xctl_phase(t_paf4xctl *x, float mainphase, float shiftphase,
    float vibphase)
{
    x->x_phase = mainphase;
    x->x_shiftphase = shiftphase;
    x->x_vibphase = vibphase;
    x->x_triggerme = 1;
}

    /* value of Cauchy distribution at TABRANGE */
#define CAUCHYVAL (1./ (1. + TABRANGE * TABRANGE))
    /* first derivative of Cauchy distribution at TABRANGE */
#define CAUCHYSLOPE ((-2. * TABRANGE) * CAUCHYVAL * CAUCHYVAL)
#define ADDSQ (- CAUCHYSLOPE / (2 * TABRANGE))

static void paf4x_dosetup(void)
{
    int i;
    float CAUCHYFAKEAT3  =
    	(CAUCHYVAL + ADDSQ * TABRANGE * TABRANGE);
    float CAUCHYRESIZE = (1./ (1. - CAUCHYFAKEAT3));
    for (i = 0; i <= TABSIZE; i++)
    {
    	float f = i * ((float)TABRANGE/(float)TABSIZE);
	float gauss = exp(-f * f);
	float cauchygenuine = 1. / (1. + f * f);
	float cauchyfake = cauchygenuine + ADDSQ * f * f;
	float cauchyrenorm = (cauchyfake - 1.) * CAUCHYRESIZE + 1.;
	if (i != TABSIZE)
	{
	    paf4x_gauss[i].p_y = gauss;
	    paf4x_cauchy[i].p_y = cauchyrenorm;
	    /* post("%f", cauchyrenorm); */
	}
	if (i != 0)
	{
	    paf4x_gauss[i-1].p_diff = gauss - paf4x_gauss[i-1].p_y;
	    paf4x_cauchy[i-1].p_diff = cauchyrenorm - paf4x_cauchy[i-1].p_y;
	}
    }
}

#ifdef TESTME

#define BS 64
main()
{
    t_paf4xctl x;
    float x1[BS];
    int i;
    paf4x_dosetup();
    paf4xctl_init(&x);
    paf4xctl_setsr(&x, 16000., BS);
    paf4xctl_freq(&x, 1000, 0);
    paf4xctl_bw(&x, 2000, 0);
    paf4xctl_amp(&x, 1000, 0);
    paf4xctl_run(&x, x1, BS);
    for (i = 0; i < BS/4; i++)
    {
	printf("%15.5f %15.5f %15.5f %15.5f\n",
		x1[4*i], x1[4*i+1], x1[4*i+2], x1[4*i+3]);
    }
#if 0
    printf("\n");
    paf4xctl_bw(&x, 2000, 0);
    paf4xctl_run(&x, x1, BS);
    for (i = 0; i < BS/4; i++)
    {
	printf("%15.5f %15.5f %15.5f %15.5f\n",
		x1[4*i], x1[4*i+1], x1[4*i+2], x1[4*i+3]);
    }
#endif
}

#endif
#ifdef FTS1X

static fts_symbol_t *paf4x_dspname;

typedef struct _paf4x
{
  fts_object_t ob;		/* object header */
  t_paf4xctl paf4xctl;
} paf4x_t;

/* ---------------------------------------- */
/* Methods                                  */
/* ---------------------------------------- */

    /* formalities... */

static void paf4x_dspfun(fts_word_t *a)
{
    paf4x_t *x = (paf4x_t *)fts_word_get_obj(a);
    float *out1 = (float *)fts_word_get_obj(a + 1);
    long n = fts_word_get_long(a + 3);
    
    paf4xctl_run(&x->paf4xctl, out1, n);
}

void paf4x_put(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *x = (paf4x_t *)o;
    fts_dsp_descr_t *dsp = (fts_dsp_descr_t *)fts_get_obj_arg(at, ac, 0, 0);
    fts_atom_t a[4];
    float sr = fts_dsp_get_output_srate(dsp, 0);
    int vecsize = fts_dsp_get_output_size(dsp, 0);

    paf4xctl_setsr(&x->paf4xctl, sr, vecsize);
    fts_set_obj(a, x);
    fts_set_symbol(a+1, fts_dsp_get_output_name(dsp, 0));
    fts_set_long(a+2, fts_dsp_get_output_size(dsp, 0));
    dsp_add_funcall(paf4x_dspname, 3, a);
}

static void paf4x_freq(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    float val = fts_get_float_long_arg( at, ac, 0, 0.0f);
    int time = fts_get_long_arg(at, ac, 1, 0);
    paf4xctl_freq(&this->paf4xctl, val, time);
}

static void paf4x_cf(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    float val = fts_get_float_long_arg( at, ac, 0, 0.0f);
    int time = fts_get_long_arg(at, ac, 1, 0);
    paf4xctl_cf(&this->paf4xctl, val, time);
}

static void paf4x_bw(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    float val = fts_get_float_long_arg( at, ac, 0, 0.0f);
    int time = fts_get_long_arg(at, ac, 1, 0);
    paf4xctl_bw(&this->paf4xctl, val, time);
}

static void paf4x_amp(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    
    float val = fts_get_float_long_arg( at, ac, 0, 0.0f);
    int time = fts_get_long_arg(at, ac, 1, 0);
    paf4xctl_amp(&this->paf4xctl, val, time);
}

static void paf4x_vib(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    float val = fts_get_float_long_arg( at, ac, 0, 0.0f);
    int time = fts_get_long_arg(at, ac, 1, 0);
    paf4xctl_vib(&this->paf4xctl, val, time);
}

static void paf4x_vfr(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    float val = fts_get_float_long_arg( at, ac, 0, 0.0f);
    int time = fts_get_long_arg(at, ac, 1, 0);
    paf4xctl_vfr(&this->paf4xctl, val, time);
}

static void paf4x_shift(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    float val = fts_get_float_long_arg( at, ac, 0, 0.0f);
    int time = fts_get_long_arg(at, ac, 1, 0);
    paf4xctl_shift(&this->paf4xctl, val, time);
}

static void paf4x_phase(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    float phase = fts_get_float_long_arg( at, ac, 0, 0.0f);
    float shiftphase = fts_get_long_arg(at, ac, 1, 0);
    float vibphase = fts_get_long_arg(at, ac, 2, 0);
    paf4xctl_shift(&this->paf4xctl, phase, shiftphase, vibphase);
}

static void
paf4x_print(fts_object_t *o, int winlet, fts_symbol_t *s, int ac, const fts_atom_t *at)
{
}

static void
paf4x_init(fts_object_t *o, int winlet, fts_symbol_t *s, int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;
    int i;

    dsp_list_insert(o); 	/* put object in list */
    paf4xctl_init(&this->paf4xctl);
}

static void paf4x_delete(fts_object_t *o, int winlet, fts_symbol_t *s,
    int ac, const fts_atom_t *at)
{
    paf4x_t *this = (paf4x_t *)o;

    dsp_list_remove(o);
}

static fts_status_t paf4x_instantiate( fts_class_t *cl,
    int ac, const fts_atom_t *at)
{
    fts_atom_type_t a[3];
    
    fts_class_init( cl, sizeof(paf4x_t), 1, 2, 0);
    
    a[0] = fts_Symbol;
    fts_method_define( cl, fts_SystemInlet, fts_s_init, paf4x_init, 1, a);
    fts_method_define( cl, fts_SystemInlet, fts_s_delete, paf4x_delete, 1, a);
    
    a[0] = fts_Float|fts_Long;
    a[1] = fts_Long;
    fts_method_define( cl, 0, fts_new_symbol("freq"), paf4x_freq, 2, a);
    fts_method_define( cl, 0, fts_new_symbol("cf"), paf4x_cf, 2, a);
    fts_method_define( cl, 0, fts_new_symbol("bw"), paf4x_bw, 2, a);
    fts_method_define( cl, 0, fts_new_symbol("amp"), paf4x_amp, 2, a);
    fts_method_define( cl, 0, fts_new_symbol("vib"), paf4x_vib, 2, a);
    fts_method_define( cl, 0, fts_new_symbol("vfr"), paf4x_vfr, 2, a);
    fts_method_define( cl, 0, fts_new_symbol("shift"), paf4x_shift, 2, a);
    fts_method_define( cl, 0, fts_new_symbol("phase"), paf4x_phase, 3, a);
    
    fts_method_define( cl, 0, fts_new_symbol("print"), paf4x_print, 0, 0);
    
    a[0] = fts_Object;  
    fts_method_define(cl, fts_SystemInlet, fts_new_symbol("put"),
	paf4x_put, 1, a);
    
    dsp_sig_inlet(cl, 0);	/* order forcing only */
    
    dsp_sig_outlet(cl, 0);
    
    return fts_Success;
}

fts_symbol_t *paf4x_qui;

void paf4x_config(void)
{
    sys_log(paf4x_version);
    paf4x_dspname = fts_new_symbol("paf4x");
    dsp_declare_function(paf4x_dspname, paf4x_dspfun);
    fts_metaclass_create(fts_new_symbol("paf4x"),
	paf4x_instantiate, fts_always_equiv);
    paf4x_dosetup();
}

fts_module_t paf4x_module =
  {"paf4x", "paf4x", paf4x_config, 0};

#endif /* FTS1X */

#ifdef V26

typedef struct _paf4x
{
    t_head x_h;
    t_sig *x_io[IN1+OUT1];
    t_paf4xctl x_paf4xctl;
} t_paf4x;

static void paf4x_put(t_paf4x *x, long int whether)
{
    if (whether)
    {
	float sr = x->x_io[0]->s_sr;
	int vecsize = x->x_io[0]->s_n;
    	u_stdout((t_ugen *)x);
	paf4xctl_setsr(&x->x_paf4xctl, sr, vecsize);
	dspchain_addc(paf4xctl_run, 3,
	    &x->x_paf4xctl, x->x_io[1]->s_shit, x->x_io[1]->s_n);
    }
}

static void paf4x_freq(t_paf4x *x, double val, int time)
{
    paf4xctl_freq(&x->x_paf4xctl, val, time);
}

static void paf4x_cf(t_paf4x *x, double val, int time)
{
    paf4xctl_cf(&x->x_paf4xctl, val, time);
}

static void paf4x_bw(t_paf4x *x, double val, int time)
{
    paf4xctl_bw(&x->x_paf4xctl, val, time);
}

static void paf4x_amp(t_paf4x *x, double val, int time)
{
    paf4xctl_amp(&x->x_paf4xctl, val, time);
}

static void paf4x_vib(t_paf4x *x, double val, int time)
{
    paf4xctl_vib(&x->x_paf4xctl, val, time);
}

static void paf4x_vfr(t_paf4x *x, double val, int time)
{
    paf4xctl_vfr(&x->x_paf4xctl, val, time);
}

static void paf4x_shift(t_paf4x *x, double val, int time)
{
    paf4xctl_shift(&x->x_paf4xctl, val, time);
}

static void paf4x_phase(t_paf4x *x, double phase, double shiftphase,
    double vibphase)
{
    paf4xctl_phase(&x->x_paf4xctl, phase, shiftphase, vibphase);
}

static void paf4x_debug(t_paf4x *x)
{
    linenv_debug( &x->x_paf4xctl.x_ampenv, "amp");
    linenv_debug(&x->x_paf4xctl.x_freqenv, "fre");
}

t_externclass *paf4x_class;

static void *paf4x_new()
{
    t_paf4x *x = (t_paf4x *)obj_new(&paf4x_class, 0);
    u_setup((t_ugen *)x, IN1, OUT1);
    u_setforcer(x);
    paf4xctl_init(&x->x_paf4xctl);
    return (x);
}

void sigpaf4x_setup()
{
    post(paf4x_version);
    c_extern(&paf4x_class, paf4x_new, u_clean,
	gensym("paf4x"), sizeof(t_paf4x), 0,  0);
    c_addmess(paf4x_put, gensym("put"), A_CANT, 0);
    c_addmess(paf4x_freq, gensym("freq"), A_FLOAT, A_LONG, 0);
    c_addmess(paf4x_cf, gensym("cf"), A_FLOAT, A_LONG, 0);
    c_addmess(paf4x_bw, gensym("bw"), A_FLOAT, A_LONG, 0);
    c_addmess(paf4x_amp, gensym("amp"), A_FLOAT, A_LONG, 0);
    c_addmess(paf4x_vib, gensym("vib"), A_FLOAT, A_LONG, 0);
    c_addmess(paf4x_vfr, gensym("vfr"), A_FLOAT, A_LONG, 0);
    c_addmess(paf4x_shift, gensym("shift"), A_FLOAT, A_LONG, 0);
    c_addmess(paf4x_phase, gensym("phase"), A_FLOAT, A_FLOAT, A_FLOAT, 0);
    c_addmess(paf4x_debug, gensym("debug"), 0);
    u_inletmethod(0);
    paf4x_dosetup();
}

#endif /* V26 */

#ifdef PD

typedef struct _paf4x
{
    t_object x_obj;
    t_paf4xctl x_paf4xctl;
} t_paf4x;

static t_class *paf4x_class;

static void *paf4x_new(void)
{
    t_paf4x *x = (t_paf4x *)pd_new(paf4x_class);
    paf4xctl_init(&x->x_paf4xctl);
    outlet_new(&x->x_obj, gensym("signal"));
    return (x);
}

static t_int *paf4x_perform(t_int *w)
{
    t_paf4xctl *ctl = (t_paf4xctl *)(w[1]);
    t_float *out1 = (t_float *)(w[2]);
    int n = (int)(w[3]);
    paf4xctl_run(ctl, out1, n);
    return (w+4);
}

static void paf4x_dsp(t_paf4x *x, t_signal **sp)
{
    float sr = sp[0]->s_sr;
    int vecsize = sp[0]->s_n;
    paf4xctl_setsr(&x->x_paf4xctl, sr, vecsize);
    dsp_add(paf4x_perform, 3,
	&x->x_paf4xctl, sp[0]->s_vec, sp[0]->s_n);
}

static void paf4x_freq(t_paf4x *x, t_floatarg val, t_floatarg time)
{
    paf4xctl_freq(&x->x_paf4xctl, val, time);
}

static void paf4x_cf(t_paf4x *x, t_floatarg val, t_floatarg time)
{
    paf4xctl_cf(&x->x_paf4xctl, val, time);
}

static void paf4x_bw(t_paf4x *x, t_floatarg val, t_floatarg time)
{
    paf4xctl_bw(&x->x_paf4xctl, val, time);
}

static void paf4x_amp(t_paf4x *x, t_floatarg val, t_floatarg time)
{
    paf4xctl_amp(&x->x_paf4xctl, val, time);
}

static void paf4x_vib(t_paf4x *x, t_floatarg val, t_floatarg time)
{
    paf4xctl_vib(&x->x_paf4xctl, val, time);
}

static void paf4x_vfr(t_paf4x *x, t_floatarg val, t_floatarg time)
{
    paf4xctl_vfr(&x->x_paf4xctl, val, time);
}

static void paf4x_shift(t_paf4x *x, t_floatarg val, t_floatarg time)
{
    paf4xctl_shift(&x->x_paf4xctl, val, time);
}

static void paf4x_phase(t_paf4x *x, t_floatarg mainphase, t_floatarg shiftphase,
    t_floatarg vibphase)
{
    paf4xctl_phase(&x->x_paf4xctl, mainphase, shiftphase, vibphase);
}

static void paf4x_setcauchy(t_paf4x *x, t_floatarg f)
{
    x->x_paf4xctl.x_cauchy = (f != 0);
}

static void paf4x_set4x(t_paf4x *x, t_floatarg f)
{
    x->x_paf4xctl.x_4xcompat = (f != 0);
}

static void paf4x_debug(t_paf4x *x)
{
    /* whatever you want... */
}

void paf4x_tilde_setup(void)
{
    post(paf4x_version);
    paf4x_class = class_new(gensym("paf4x~"), (t_newmethod)paf4x_new, 0,
    	sizeof(t_paf4x), 0, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_dsp, gensym("dsp"), A_CANT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_freq, gensym("freq"),
    	A_FLOAT, A_DEFFLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_cf, gensym("cf"),
    	A_FLOAT, A_DEFFLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_bw, gensym("bw"),
    	A_FLOAT, A_DEFFLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_amp, gensym("amp"),
    	A_FLOAT, A_DEFFLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_vib, gensym("vib"),
    	A_FLOAT, A_DEFFLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_vfr, gensym("vfr"),
    	A_FLOAT, A_DEFFLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_shift, gensym("shift"),
    	A_FLOAT, A_DEFFLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_phase, gensym("phase"),
    	A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_setcauchy, gensym("cauchy"),
    	A_FLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_set4x, gensym("4x"),
    	A_FLOAT, 0);
    class_addmethod(paf4x_class, (t_method)paf4x_debug, gensym("debug"), 0);
    paf4x_dosetup();
}

#endif /* PD */
