/* A FTS extern object to follow an EXPLODE */
/* By : Francois Dechelle and Miller Puckette */

/* follow from a settable (by name) instance of explode                 */
/* Accepts the messages :                                               */
/*  follow : start                                                      */
/*  followat (int pitch, int velocity, int chan) : start at note        */
/*  stop : ???                                                          */
/*  <int> : incoming time ??                                            */
/*  <in1> : incoming pitch                                              */
/*  <in2> : incoming velocity                                           */

#define MAX_EXTERN
#include <m_extern.h>
#include <x_explode.h>

/* ---------- the voicefollow instance structure ---------- */
typedef struct S_voicefollow {
  t_object ob;               /* object header */
  void *outlet_0;            /* outlet */
  void *outlet_1;            /* outlet */
  void *outlet_2;            /* outlet */
  void *outlet_3;            /* outlet */
  void *outlet_4;            /* outlet */
  /* ---------- */
  enum { FOLLOWING, STOPPED} mode;
  t_symbol *sym;           /* to connect to an explode object */
  t_explode *explode;      /* the explode object to which is connected */
  t_evt *current;
  int setpitch, setvelocity;
  long incomingpitch, incomingvelocity;
  int setduration, setchan;
  long incomingduration, incomingchan;
	int lastpitmatched;
	int algorithm;
} voicefollow;


/* ---------- method print : for debug ---------- */
void voicefollow_print( x)
     voicefollow *x;
{
  post( "(voicefollow *)0x%x ={", (long)x);
  post( "%s, /* mode */", (x->mode==FOLLOWING) ? "FOLLOWING" : "STOPPED");
}

/* ---------- auxiliary ---------- */
t_symbol *voicefollow_explodesym;

t_explode *getexplodebyname( t_symbol *s)
{
  t_explode *tmp;

  tmp = (t_explode *) (s->s_thing);
  if ( (!tmp) || (obj_getclassname(tmp) != voicefollow_explodesym) )
    tmp = (t_explode *)0;
  return tmp;
}

/* ---------- method stop : stop and reinitialize ---------- */
void voicefollow_stop(x)
     voicefollow *x;
{
  if ( x->mode == FOLLOWING)
    {
      x->setpitch = x->setvelocity = 0;
      x->mode = STOPPED;
    }
}

/* ---------- aux method : send a matched event ---------- */
voicefollow_sendevt(x, e, noteon)
     voicefollow *x;
     t_evt *e;
     int noteon;
{
  outlet_int(x->outlet_4, e->e_chan);
  outlet_int(x->outlet_3, e->e_dur);
  outlet_int(x->outlet_2, noteon ? e->e_vel : 0L);
  outlet_int(x->outlet_1, e->e_pit);
}

#define AHEAD 3


/* ---------- score following algorithm number 1 ---------- */
voicefollow_noteon( x, n)
     voicefollow *x; long n;
{
	int pits[AHEAD], i, matchpit;
	t_evt *e;
	for (e = x->current, i = 0; e && i < AHEAD; e = e->e_next)
	{
		if (e->e_chan == 1) pits[i++] = e->e_pit;
	}
	while (i < AHEAD) pits[i++] = 0;

	if (n < 100 && (n == x->lastpitmatched || n == x->lastpitmatched+1 ||
		n == x->lastpitmatched-1) && n != pits[0]) return;
	
	if (n == pits[0] || n == pits[0] + 1 || n == pits[0] - 1)
	{
		matchpit = pits[0];
		goto matched;
	}
	for (i = 1; i < AHEAD; i++) if (n == pits[i])
	{
		matchpit = n;
		goto matched;
	}
	return;
matched:
	
	for (e = x->current; e; e = e->e_next)
	{
		if (e->e_chan == 1 && e->e_pit == matchpit) break;
		else if (e->e_chan != 1) voicefollow_sendevt(x, e, 1);
	}
	if (!e)
	{
		post("voicefollow bug 1");
		voicefollow_stop(x);
		return;
	}
		
  /* output the winning note */
  else voicefollow_sendevt(x, e, 1);
	x->lastpitmatched = matchpit;
  /* and set pointer to the note after that */
  x->current = e->e_next;
}

/* ---------- score following algorithm number 2 ---------- */
voicefollow_noteon2( x, n)
     voicefollow *x; long n;
{
	int pits[AHEAD], i, matchpit;
	t_evt *e;
	for (e = x->current, i = 0; e && i < AHEAD; e = e->e_next)
	{
		if (e->e_chan == 1) pits[i++] = e->e_pit;
	}
	while (i < AHEAD) pits[i++] = 0;

	if (n == pits[0])
	{
		matchpit = n;
		goto matched;
	}	
	if (n < 100 && (n == x->lastpitmatched || n == x->lastpitmatched+1 ||
		n == x->lastpitmatched-1)) return;	
	if (n == pits[0] + 1 || n == pits[0] - 1)
	{
		matchpit = pits[0];
		goto matched;
	}
	if (n == pits[1])
	{
		matchpit = n;
		goto matched;
	}
	if (n < 100 && (n == x->lastpitmatched-2 || n == x->lastpitmatched+2))
		return;	
	if (n == pits[0] + 2 || n == pits[0] - 2)
	{
		matchpit = pits[0];
		goto matched;
	}
	if (n == pits[1] + 1 || n == pits[1] - 1)
	{
		matchpit = pits[1];
		goto matched;
	}
	if (n == pits[2])
	{
		matchpit = n;
		goto matched;
	}
	return;

matched:
	
	for (e = x->current; e; e = e->e_next)
	{
		if (e->e_chan == 1 && e->e_pit == matchpit) break;
		else if (e->e_chan != 1) voicefollow_sendevt(x, e, 1);
	}
	if (!e)
	{
		post("voicefollow bug 1");
		voicefollow_stop(x);
		return;
	}
		
  /* output the winning note */
  else voicefollow_sendevt(x, e, 1);
	x->lastpitmatched = matchpit;
  /* and set pointer to the note after that */
  x->current = e->e_next;
}

/* ---------- method int : process an incoming pitch ---------- */
void voicefollow_int( x, time)
     voicefollow *x;
     long time;
{
  if ( x->mode != FOLLOWING)
    return;
  if ( (x->setvelocity && x->incomingvelocity) || !x->setvelocity)
	{	
		if (x->algorithm == 1) voicefollow_noteon( x, x->incomingpitch);
		else if (x->algorithm == 2) voicefollow_noteon2( x, x->incomingpitch);
	}
}

/* ---------- method in1 : get an incoming pitch on inlet 1 ---------- */
void voicefollow_in1( x, pitch)
     voicefollow *x;
     long pitch;
{
  x->setpitch = 1;
  x->incomingpitch = pitch;
}

/* ---------- aux method connecttoexplode : ---------- */
void voicefollow_connecttoexplode( x)
t_voicefollow *x;
{
  t_explode *y;

  y = getexplodebyname( x->sym);
  if ( !y )
    {
      error( "voicefollow : %s is not an explode", x->sym->s_name);
      x->current = 0;
      return;
    }
  if ( y->x_next)
    warn( "voicefollow : %s is multiply defined", x->sym->s_name);
  if ( !y->x_serial )
    {
      error( "voicefollow : %s is recording", x->sym->s_name);
      x->current = 0;
      return;
    }
  x->current = y->x_evt;
}

/* ---------- method follow : start following ---------- */
void voicefollow_follow(x, n)
t_voicefollow *x; long n;
{
  t_evt *e;

  voicefollow_stop( x);
  voicefollow_connecttoexplode( x);
  
  for (e = x->current; e && n--; e = e->e_next)
    ;
  if ( !e )
    post("voicefollow: out of range");
  x->current = e;
	x->lastpitmatched = 0;
  x->mode = FOLLOWING;
}

/* ---------- method followat : ... ---------- */
void voicefollow_followat( x, pitch, velocity, chan)
t_voicefollow *x; long pitch, velocity, chan;
{
  t_evt *e;

  voicefollow_stop(x);
  voicefollow_connecttoexplode( x);

  for (e = x->current; e; e = e->e_next)
    if (e->e_pit == pitch && e->e_vel == velocity && e->e_chan == chan)
      break;
  if (!e)
    post("voicefollow: note not found");
  x->current = (e) ? e->e_next : 0;
	x->lastpitmatched = 0;

  x->mode = FOLLOWING;
}

/* ---------- method params : set the params for score following ---------- */
void voicefollow_params(x, nfwd, ftime, oct, tolerance)
t_voicefollow *x;
     long nfwd, ftime, oct, tolerance;
{
}

/* ---------- method free : free the object ---------- */
void voicefollow_free( x)
t_voicefollow *x;
{
  /* should free the hang and skip lists */
}

/* ---------- pointer to the class ---------- */
t_externclass *voicefollow_class;

/* ---------- method new : instance creation ---------- */
void *voicefollow_new(s)
     t_symbol *s;
{
  voicefollow *x;
  t_explode *y;

  x = (voicefollow *) obj_new( &voicefollow_class, 0);

  /* create outlet, declaring it to send "int" */
  x->x_outlet5 = outlet_new(x, gensym("int"));
  x->x_outlet4 = outlet_new(x, gensym("int"));
  x->x_outlet3 = outlet_new(x, gensym("int"));
  x->x_outlet2 = outlet_new(x, gensym("int"));
  x->x_outlet1 = outlet_new(x, gensym("int"));

  /* create inlets */
  inlet_new(x, x, gensym("float"), gensym("ft1")); /* provisional pitch */

  x->sym = s;   /* initialise symbol to connect to explode object */

  x->mode = STOPPED;
  return x;
}

/* ---------- method setup : class creation ---------- */
voicefollow_setup()
{
  c_extern( &voicefollow_class, voicefollow_new, voicefollow_free,
		gensym("voicefollow"),
			sizeof(voicefollow), 0, A_SYM, 0);

  c_addfloat( voicefollow_float);

  c_addmess( voicefollow_follow, gensym("follow"), A_DEFLONG, 0);
  c_addmess( voicefollow_followat, gensym("followat"),
		A_LONG, A_LONG, A_LONG, 0);

  c_addmess( voicefollow_stop, gensym("stop"), 0);

  c_addmess( voicefollow_ft1, gensym( "ft1"), A_FLOAT, 0);

  c_addmess( voicefollow_print, gensym( "print"), 0);
  c_addmess( voicefollow_params, gensym( "params"), A_LONG, A_LONG, A_LONG, 0);

  voicefollow_explodesym = gensym( "explode");
}

