// r16 - source
// FTS by Miller Puckette; copyright 1990 ircam

/*	sigpt__makespect().  Support routine for pitch tracker.

void sigpt__makespect(fftbuf, kp, spect, n)
 float *fftbuf, *spect;
 t_kernel *kp;
 int n;

//with:

typedef struct kernel
{
	float *k_coef;
	int k_ncoef;
	int k_onset;
	float k_powcorrect;
} t_kernel;

dots the real vector in k_coef with the complex fft in fftbuf (giving
a complex result) and puts the square modulus of the result in the
array spect.  k_ncoef gives the length of the dot product and k_onset is
an onset into the fft, in floats (i.e. twice the index if fftbuf is viewed
as a complex array.)

*/


#define k_coef		0
#define k_ncoef		4
#define k_onset		8

//integers -- arguments
#define IFFT		r16
#define IKERNEL		r17
#define ISPECT		r18
#define INPOINT		r19

//temporaries
#define ICOEFCOUNT	r20
#define IFFTPOINT	r21
#define ICOEFPOINT	r22
#define ICOEFDECR		r23

//single precision FP
#define SUMR	f8
#define SUMI	f9
#define COEFA	f10
#define COEFB	f11
#define FFTRA	f12
#define FFTIA	f13
#define FFTRB	f14
#define FFTIB	f15

.data
.align 4
.text
.align 8
_sigpitch__makespect::
	adds		-1, INPOINT, INPOINT
	bc			alldone
	adds		-1, r0, ICOEFDECR
	ld.l		k_onset(IKERNEL), r31	// offset into FFT
	shl			2, r31, r31
	adds		IFFT, r31, IFFTPOINT		// points to first FFT point
	ld.l		k_coef(IKERNEL), ICOEFPOINT	// point to coef vector
	ld.l		k_ncoef(IKERNEL), ICOEFCOUNT	// number of points
	adds		-1, ICOEFCOUNT, ICOEFCOUNT	// back up one for bla
	adds		16, IKERNEL, IKERNEL		// next kernel
	
.align 8
.dual
	r2apt.ss	f0, f0, f0				// clear KI
		fld.l	0(ICOEFPOINT), COEFA	// get first coef
	ra1p2.ss	f0, f0, f0				// this 3 times to clear pipeline
		fld.d	0(IFFTPOINT), FFTRA		// get first FFT point
	ra1p2.ss	f0, f0, f0
		bla		ICOEFDECR, ICOEFCOUNT, loop
	ra1p2.ss	f0, f0, f0
		fld.l	4(ICOEFPOINT)++, COEFB
loop:
// ******************* LOOP ***********************
	m12apm.ss	COEFA, FFTRA, f0
		fld.d	8(IFFTPOINT)++, FFTRB
	m12apm.ss	COEFA, FFTIA, f0
		bla		ICOEFDECR, ICOEFCOUNT, loop2
	m12apm.ss	f0, f0, f0
		fld.l	4(ICOEFPOINT)++, COEFA
// ************** END OF LOOP *********************
.enddual
	m12apm.ss	f0, f0, f0	// fall out of loop here if odd # points
		nop
	m12apm.ss	f0, f0, f0
		nop
	br doneloop
	nop

loop2:
.dual
// ******************* LOOP2 ***********************
	m12apm.ss	COEFB, FFTRB, f0
		fld.d	8(IFFTPOINT)++, FFTRA
	m12apm.ss	COEFB, FFTIB, f0
		bla		ICOEFDECR, ICOEFCOUNT, loop
	m12apm.ss	f0, f0, f0
		fld.l	4(ICOEFPOINT)++, COEFB
// ************** END OF LOOP *********************
.enddual
	m12apm.ss	f0, f0, f0	// fall out of loop here if even # points
		nop
	m12apm.ss	f0, f0, f0
		nop

doneloop:
	m12apm.ss	f0, f0, f0
	m12apm.ss	f0, f0, SUMR
	m12apm.ss	f0, f0, SUMI
	fmul.ss		SUMR, SUMR, f30
	fmul.ss		SUMI, SUMI, f31
	fadd.ss		f30, f31, f31
	fst.l		f31, 0(ISPECT)
	
	br _sigpitch__makespect
	adds		4, ISPECT, ISPECT

alldone:
	bri	r1
	nop






// r17 - dest
// r18 - npoints
// r19 - decrement for "bla" instruction
// f8-f12 first bank of 4 registers
// f13-f16 second bank


.text


.align 8
_sigpitch_sigrsqrt1::
.dual
	fld.q		0(r16), f12
	adds		-4, r0, r19
	frsqr.ss		f12, f20
		adds		-4, r18, r18
	frsqr.ss		f13, f21
		fld.q		16(r16)++, f8
	frsqr.ss		f14, f22
		bla			r19, r18, loop11
	frsqr.ss		f15, f23
		adds		-16, r17, r17

// ******************* LOOP 1 ***********************
loop11:
	frsqr.ss		f8, f16
		nop
	frsqr.ss		f9, f17
		fld.q	16(r16)++, f12
	frsqr.ss		f10, f18
		bla		r19, r18, loop112
	frsqr.ss		f11, f19
		fst.q	f20, 16(r17)++

// ************* can fall out of loop here *************
.enddual
	fnop
		bri r1
	fnop
		nop

.dual
// *************** loop continues ******************
loop112:
	frsqr.ss		f12, f20
		nop
	frsqr.ss		f13, f21
		fld.q	16(r16)++, f8
	frsqr.ss		f14, f22
		bla		r19, r18, loop11
	frsqr.ss		f15, f23
		fst.q	f16, 16(r17)++
// ************** END OF LOOP *********************
.enddual
	fnop
		bri r1
	fnop
		nop


// --------------- multiplex ------------
#define MUXIN	r16
#define MUXOUT r17
#define MUXSPREAD	r18
#define MUXNPOINTS	r19
#define MUXDECR r20

.text
_sigpitch_mux::
	mov		-4, MUXDECR
	adds	-8, MUXOUT, MUXOUT
	fld.q	0(MUXIN), f16
	adds	MUXDECR, MUXNPOINTS, MUXNPOINTS
	bla		MUXDECR, MUXNPOINTS, muxloop
	adds	8, MUXIN, MUXIN
muxloop:
	fst.l	f16, 8(MUXOUT)++
	fst.l	f17, 8(MUXOUT)++
	fld.d	8(MUXIN)++, f16
	fst.l	f18, 8(MUXOUT)++
	fst.l	f19, 8(MUXOUT)++
	bla		MUXDECR, MUXNPOINTS, muxloop
	fld.d	8(MUXIN)++, f18

	bri	r1
	nop

// ---------------- copy ------------


// integers -- arguments
#define INADDR		r16
#define IOUTADDR	r17
#define ISIZE		r18
//temporaries
#define IDECR		r19
#define IN2			r20
#define ISIZE2		r21

.text
.align 4
	nop
_sigpitch_copy::
	fld.q	0(INADDR), f16
	adds	-16, r0, IDECR
	adds	-16, ISIZE, ISIZE
	adds	-16, IOUTADDR, IOUTADDR
	bla		IDECR, ISIZE, loop22
	fld.q	16(INADDR)++, f20
loop22:
	fst.q	f16, 16(IOUTADDR)++
	bla		IDECR, ISIZE, loop23
	fld.q	16(INADDR)++, f16

	bri		r1
	nop	

loop23:
	fst.q	f20, 16(IOUTADDR)++
	bla		IDECR, ISIZE, loop22
	fld.q	16(INADDR)++, f20

	bri		r1
	nop	

